/*
 * Graphic lambda calculus in WebGL. This is a clone of the "GCalc" program
 * included in the "plt-games" package from the Racket Scheme distribution.
 * The original GCalc program was written by Eli Barzilay.
 *
 * Copyright (c) 2017 Pim Goossens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * expr ::= ["v", color]       // Variable
 *        | ["|", expr, expr]  // Left/right subdivision
 *        | ["-", expr, expr]  // Top/bottom
 *        | ["/", expr, expr]  // Front/back
 *        | ["l", color, expr] // Abstraction (lambda/function creation)
 *        | ["a", expr, expr]  // Application (function call)
 */

"use strict";

function mkrgb()
{
	if (arguments.length == 3)
		return ["v", Array.from(arguments)];
	else if (arguments.length == 1)
		return ["v", arguments[0]];
	else
		return ["v", null];
}

var grid =
{
	main: mkrgb(null),

	color: function(r, g, b)
	{
		this.main = mkrgb(r, g, b);
	},
	set:    function(expr)  { this.main = expr; },
	left:   function(expr)  { this.set(["|", expr, this.main]); },
	right:  function(expr)  { this.set(["|", this.main, expr]); },
	over:   function(expr)  { this.set(["-", expr, this.main]); },
	under:  function(expr)  { this.set(["-", this.main, expr]); },
	back:   function(expr)  { this.set(["/", expr, this.main]); },
	front:  function(expr)  { this.set(["/", this.main, expr]); },
	applyf: function(expr)  { this.set(["a", expr, this.main]); },
	applya: function(expr)  { this.set(["a", this.main, expr]); },
	lambda: function(color) { this.set(["l", color, this.main]); },
	eval:   function()      { this.set(evalgc(this.main)); }
};

function lerp(a, b, t) { return a*(1.0-t) + b*t; }
function avg(a, b)     { return (a+b)/2.0;       }

// db = block drawing function
// da = abstraction drawing function
function drawexpr(expr, db, da, x1=0.0, y1=0.0, z1=0.0, x2=1.0, y2=1.0, z2=1.0)
{
	if (expr == null || expr.length == 0)
		return;

	switch (expr[0])
	{
		case "v":
			db(expr[1], x1, y1, z1, x2, y2, z2);
			break;
		case "|":
			drawexpr(expr[1], db, da, x1, y1, z1, avg(x1, x2), y2, z2);
			drawexpr(expr[2], db, da, avg(x1, x2), y1, z1, x2, y2, z2);
			break;
		case "-":
			drawexpr(expr[1], db, da, x1, avg(y1, y2), z1, x2, y2, z2);
			drawexpr(expr[2], db, da, x1, y1, z1, x2, avg(y1, y2), z2);
			break;
		case "/":
			drawexpr(expr[1], db, da, x1, y1, z1, x2, y2, avg(z1, z2));
			drawexpr(expr[2], db, da, x1, y1, avg(z1, z2), x2, y2, z2);
			break;
		case "l":
			drawexpr(expr[2], db, da,
				lerp(x1, x2, 0.1),
				lerp(y1, y2, 0.1),
				lerp(z1, z2, 0.1),
				lerp(x1, x2, 0.9),
				lerp(y1, y2, 0.9),
				lerp(z1, z2, 0.9));
			da(expr[1],
				lerp(x1, x2, 0.001),
				lerp(y1, y2, 0.001),
				lerp(z1, z2, 0.001),
				lerp(x1, x2, 0.999),
				lerp(y1, y2, 0.999),
				lerp(z1, z2, 0.999));
			break;
		case "a":
			drawexpr(expr[1], db, da,
				x1,
				lerp(y1, y2, 0.3),
				lerp(z1, z2, 0.3),
				lerp(x1, x2, 0.4),
				lerp(y1, y2, 0.7),
				lerp(z1, z2, 0.7));
			drawexpr(expr[2], db, da,
				lerp(x1, x2, 0.6),
				lerp(y1, y2, 0.3),
				lerp(z1, z2, 0.3),
				x2,
				lerp(y1, y2, 0.7),
				lerp(z1, z2, 0.7));
			break;
	}
}
