/*
 * Graphic lambda calculus in WebGL. This is a clone of the "GCalc" program
 * included in the "plt-games" package from the Racket Scheme distribution.
 * The original GCalc program was written by Eli Barzilay.
 *
 * Copyright (c) 2017 Pim Goossens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

"use strict";

const VERT_SIZE = 3;

const CUBE_VERTS =
[
	1, 1, 1,
	0, 1, 1,
	0, 0, 1,
	1, 0, 1,
	1, 1, 0,
	0, 1, 0,
	0, 0, 0,
	1, 0, 0
];

const CUBE_INDICES =
[
	0, 1, 2,
	2, 3, 0,
	0, 3, 7,
	7, 4, 0,
	0, 4, 5,
	5, 1, 0,
	4, 7, 6,
	6, 5, 4,
	1, 5, 6,
	6, 2, 1,
	2, 6, 7,
	7, 3, 2
];

function indexed(verts, indices)
{
	var iverts = new Array(VERT_SIZE * indices.length);
	var i = 0;

	for (let index of indices)
	{
		iverts[i++] = verts[VERT_SIZE * index];
		iverts[i++] = verts[VERT_SIZE * index + 1];
		iverts[i++] = verts[VERT_SIZE * index + 2];
	}

	return iverts;
}

// n: number of rings - 1
// d: preferred distance between vertices on a ring
function makesphere(n, d)
{
	var verts = [];

	for (var i=1; i<n; i++)
	{
		// Angular distance to vertical axis.
		var alpha = Math.PI * i / n;

		// The ring's radius.
		var r = Math.sin(alpha);

		// Position on vertical axis.
		var y = Math.cos(alpha);

		// Divide circumference by distance d to get the number of
		// vertices to place on this circle (but always at least 3).
		var nverts = Math.floor(Math.max(2*Math.PI*r / d, 3));

		for (var j=0; j<nverts; j++)
		{
			// Angle on the ring's plane.
			var beta = 2*Math.PI * j / nverts;

			var x = r*Math.cos(beta);
			var z = r*Math.sin(beta);

			verts.push(.5*x+.5, .5*y+.5, .5*z+.5);
		}
	}

	return verts;
}

//const SPHERE_VERTS = makesphere(100, .1);

function d2r(degrees)
{
	return degrees * (Math.PI / 180.0);
}

var rotatex = d2r(30.0);
var rotatey = d2r(30.0);

var rotaddx = 0.0;
var rotaddy = 0.0;

function storerotation()
{
	rotatex += rotaddx;
	rotatey += rotaddy;
	rotaddx = 0.0;
	rotaddy = 0.0;
}

function setaddedrotation(x, y)
{
	rotaddx = x / 100.0;
	rotaddy = y / 100.0;
	draw();
}

var program;
var abstprog;
var mainexpr;
var renderobj;
var abstrenderobj;

// TODO: sphere tesselation (subdivide a cube/tetrahedron?)

const VERTS = indexed(CUBE_VERTS, CUBE_INDICES);

function vsub(v1, v2)
{
	return [v2[0]-v1[0], v2[1]-v1[1], v2[2]-v1[2]];
}

function len(v)
{
	return Math.sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
}

function normalize(v)
{
	var l = len(v);
	return [v[0]/l, v[1]/l, v[2]/l];
}

function cross(v1, v2)
{
	return [
		v1[1]*v2[2] - v1[2]*v2[1],
		v1[2]*v2[0] - v1[0]*v2[2],
		v1[0]*v2[1] - v1[1]*v2[0]
	];
}

function fblock(f, verttemplate)
{
	return function(color, x1, y1, z1, x2, y2, z2)
	{
		if (color == null)
			return;

		for (var i=0; i<verttemplate.length; i+=9)
		{
			var v = new Array(3);
			for (var j=0; j<3; j++)
			{
				v[j] = new Array(3);
				v[j][0] = verttemplate[i+3*j]   * (x2 - x1) + x1;
				v[j][1] = verttemplate[i+3*j+1] * (y2 - y1) + y1;
				v[j][2] = verttemplate[i+3*j+2] * (z2 - z1) + z1;
			}

			var normal = normalize(cross(vsub(v[1], v[0]), vsub(v[2], v[0])));
			for (var j=0; j<3; j++)
			{
				f(v[j][0]);
				f(v[j][1]);
				f(v[j][2]);
				f(color[0]);
				f(color[1]);
				f(color[2]);
				f(Math.floor(normal[0] * 32767));
				f(Math.floor(normal[1] * 32767));
				f(Math.floor(normal[2] * 32767));
			}
		}
	};
}

function draw()
{
	polymode = getpolymode();
	if (grid.main !== mainexpr)
	{
		var verts = [];
		var abstverts = [];
		drawexpr(grid.main,
			fblock(v => verts.push(v), VERTS),
			fblock(v => abstverts.push(v), VERTS)
		);
		console.log(verts);
		//console.log(abstverts.length + " vertices (abst)");
		renderobj = new RenderObject(verts, null, program);
		abstrenderobj = new RenderObject(abstverts, null, abstprog, true);
		mainexpr = grid.main;
	}

	var aspect = canvaswidth / canvasheight;
	var proj = new Mat4().perspective(d2r(45.0), aspect, 1, -1);
	var view = new Mat4()
			.translate(0, 0, -2.5)
			.rotate(rotatey + rotaddy, 1, 0, 0)
			.rotate(rotatex + rotaddx, 0, 1, 0)
			.translate(-0.5, -0.5, -0.5);
	var invrot = new Mat4()
			.rotate(-rotatex - rotaddx, 0, 1, 0)
			.rotate(-rotatey - rotaddy, 1, 0, 0);

	program.uniforms.proj = proj;
	abstprog.uniforms.proj = proj;
	program.uniforms.invrot = invrot;
	abstprog.uniforms.invrot = invrot;
	program.uniforms.view = view;
	abstprog.uniforms.view = view;
	program.uniforms.model = new Mat4();
	abstprog.uniforms.model = new Mat4();

	// Clear the screen.
	gl.clearColor(0.0, 0.05, 0.15, 1.0);
	gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);

	renderobj.draw();
	abstrenderobj.draw();

	var ctx = thumbnail.getContext("2d");
	ctx.drawImage(elements.main, 0, 0, thumbnail.width, thumbnail.height);
}

function shaderchanged()
{
	updateshaders();
	updateuinfo();
	draw();
}

function initprograms()
{
	program = new Program();
	abstprog = new Program();
	// TODO
	console.log(program.sources(elements.vshader.value, elements.fshader.value));
	console.log(abstprog.sources(elements.avshader.value, elements.afshader.value));
}

function init()
{
	initelements();
	initpalette();
	initmemory();
	initgl();
	initdragdrop();
	initprograms();

	resized();
}
