/*
 * Graphic lambda calculus in WebGL. This is a clone of the "GCalc" program
 * included in the "plt-games" package from the Racket Scheme distribution.
 * The original GCalc program was written by Eli Barzilay.
 *
 * Copyright (c) 2017 Pim Goossens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

"use strict";

const DEFAULT_DEPTH_NORMAL  = 10;
const DEFAULT_DEPTH_APPL    = 10;
const DEFAULT_DEPTH_SPATIAL = 10;

function samecolor(c1, c2)
{
	if (c1 == null) return c2 == null;
	if (c2 == null) return c1 == null;
	return c1[0] == c2[0] &&
	       c1[1] == c2[1] &&
	       c1[2] == c2[2];
}

function first(type, expr)
{
	switch (expr[0])
	{
		case "|":
		case "-":
		case "/":
			if (expr[0] == type)
				return expr[1];
			else
				return [expr[0], first(type, expr[1]), first(type, expr[2])];
		default:
			return expr;
	}
}

function second(type, expr)
{
	switch (expr[0])
	{
		case "|":
		case "-":
		case "/":
			if (expr[0] == type)
				return expr[2];
			else
				return [expr[0], second(type, expr[1]), second(type, expr[2])];
		default:
			return expr;
	}
}

function evalgc(
	expr,
	dn=DEFAULT_DEPTH_NORMAL,
	da=DEFAULT_DEPTH_APPL,
	ds=DEFAULT_DEPTH_SPATIAL)
{
	if (dn <= 0)
		return expr;

	switch (expr[0])
	{
		case "a":
			switch (expr[1][0])
			{
				case "|":
					var opleft = evalgc(expr[1][1], dn, da, ds-1);
					var opright = evalgc(expr[1][2], dn, da, ds-1);
					var argleft = first("|", expr[2]);
					var argright = second("|", expr[2]);
					var left = evalgc(["a", opleft, argleft], dn, da, ds-1);
					var right = evalgc(["a", opright, argright], dn, da, ds-1);
					return ["|", left, right];
				case "-":
					var opover = evalgc(expr[1][1], dn, da, ds-1);
					var opunder = evalgc(expr[1][2], dn, da, ds-1);
					var argover = first("-", expr[2]);
					var argunder = second("-", expr[2]);
					var over = evalgc(["a", opover, argover], dn, da, ds-1);
					var under = evalgc(["a", opunder, argunder], dn, da, ds-1);
					return ["-", over, under];
				case "/":
					var opback = evalgc(expr[1][1], dn, da, ds-1);
					var opfront = evalgc(expr[1][2], dn, da, ds-1);
					var argback = first("/", expr[2]);
					var argfront = second("/", expr[2]);
					var back = evalgc(["a", opback, argback], dn, da, ds-1);
					var front = evalgc(["a", opfront, argfront], dn, da, ds-1);
					return ["/", back, front];
				case "a":
					var func = evalgc(expr[1], dn-1, da, ds);
					var args = evalgc(expr[2], dn, da-1, ds);
					return evalgc(["a", func, args], dn-1, da, ds);
				case "l":
					var color = expr[1][1];
					var body = expr[1][2];
					var arg = expr[2];

					if (da > 0)
						arg = evalgc(arg, dn, da-1, ds);

					return evalgc(subst(body, color, arg), dn-1, da, ds);
			}
		default:
			return expr;
	}
}

// Substitute _color_ for _arg_ in _expr_ wherever it's _not_ bound.
function subst(expr, color, arg)
{
	switch (expr[0])
	{
		case "v":
			if (samecolor(expr[1], color))
				return arg;
			else
				return expr;
		case "|":
			var left  = subst(expr[1], color, arg);
			var right = subst(expr[2], color, arg);
			return ["|", left, right];
		case "-":
			var over  = subst(expr[1], color, arg);
			var under = subst(expr[2], color, arg);
			return ["-", over, under];
		case "/":
			var back  = subst(expr[1], color, arg);
			var front = subst(expr[2], color, arg);
			return ["/", back, front];
		case "l":
			if (samecolor(expr[1], color))
				return expr;
			var body = subst(expr[2], color, arg);
			return ["l", expr[1], body];
		case "a":
			var func = subst(expr[1], color, arg);
			var args = subst(expr[2], color, arg);
			return ["a", func, args];
		default:
			console.warn("??? " + expr);
			return expr;
	}
}
