/*
 * Graphic lambda calculus in WebGL. This is a clone of the "GCalc" program
 * included in the "plt-games" package from the Racket Scheme distribution.
 * The original GCalc program was written by Eli Barzilay.
 *
 * Copyright (c) 2017 Pim Goossens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

"use strict";

/* For _some reason or another_, we are:
 * (a) compelled to figure out the underlying platform's
 *     byte order (see mkvbuf below), and
 * (b) required to use a hack such as the following in
 *     order to do so:
 */
const BYTE_ORDER = function()
{
	var buf = new ArrayBuffer(4);
	var dataview = new DataView(buf);
	var u32view = new Uint32Array(buf, 0, 1);
	dataview.setUint32(0, 0x12345678, false);
	return u32view[0] == 0x78563412;
}();

// Default names used for vertex attributes.
const DEFAULT_VATTR_POS = "pos";
const DEFAULT_VATTR_COLOR = "color";
const DEFAULT_VATTR_NORMAL = "normal";

/*
const ATTR =
[
	[DEFAULT_VATTR_POS, gl.FLOAT, gl.FALSE, 3],
	[DEFAULT_VATTR_COLOR, gl.BYTE, gl.TRUE, 3],
	[DEFAULT_VATTR_NORMAL, gl.SHORT, gl.TRUE, 3]
];
*/

var gl;

var polymode;

var glstate =
{
	vbuf: null,
	ibuf: null,
	vshader: null,
	fshader: null,
	prog: null
};

function mkbuf(type, arraybuf)
{
	var buf = gl.createBuffer();

	gl.bindBuffer(type, buf);
	gl.bufferData(type, arraybuf, gl.STATIC_DRAW);

	return buf;
}

function mkvbuf(data, attrinfo)
{
	var instride = attrinfo.instride;
	var outstride = attrinfo.bytestride;

	if (data.length % instride != 0)
		throw new Error("Number of data elements must be a multiple of "
			+ instride);

	var numelements = data.length / instride;
	var vbuf = new ArrayBuffer(numelements * outstride);
	var dataview = new DataView(vbuf);
	var inoffset = 0;
	var outoffset = 0;

	for (var i=0; i<numelements; i++)
	{
		for (let attr of attrinfo.attrs)
			[inoffset, outoffset] =
				attr.converter(data, dataview, inoffset, outoffset);
	}

	return [mkbuf(gl.ARRAY_BUFFER, vbuf), numelements];
}

function mkibuf(data)
{
	var maxidx = 0;
	for (let idx of data)
		maxidx = Math.max(idx, maxidx);

	var ifmt = maxidx > 255 ? gl.UNSIGNED_SHORT : gl.UNSIGNED_BYTE;

	return [mkbuf(gl.ELEMENT_ARRAY_BUFFER, Uint8Array.from(data)), ifmt];
}

function ShaderVar(index, size, type, loc)
{
	this.index = index;
	this.size = size;
	this.type = type;
	this.loc = loc;
}

function getvarinfo(prog, qname, qfunc, locfunc)
{
	var varinfo = {};

	var numvars = gl.getProgramParameter(prog, qname);
	for (var i=0; i<numvars; i++)
	{
		var info = qfunc.call(gl, prog, i);
		var loc = locfunc.call(gl, prog, info.name);
		varinfo[info.name] = new ShaderVar(i, info.size, info.type, loc);
	}

	return varinfo;
}

function getattribinfo(prog)
{
	return getvarinfo(
		prog,
		gl.ACTIVE_ATTRIBUTES,
		gl.getActiveAttrib,
		gl.getAttribLocation
	);
}

function getuniforminfo(prog)
{
	return getvarinfo(
		prog,
		gl.ACTIVE_UNIFORMS,
		gl.getActiveUniform,
		gl.getUniformLocation
	);
}

function mat2array(m)
{
	return Float32Array.from(m.transpose().values);
}

function Program()
{
	this.vshader = gl.createShader(gl.VERTEX_SHADER);
	this.fshader = gl.createShader(gl.FRAGMENT_SHADER);
	this.prog = gl.createProgram();
	this.attached = false;
	this.ainfo = null;
	this.uinfo = null;

	this.attach = function()
	{
		if (this.attached)
			return;

		gl.attachShader(this.prog, this.vshader);
		gl.attachShader(this.prog, this.fshader);

		this.attached = true;
	};

	this.detach = function()
	{
		if (!this.attached)
			return;

		gl.detachShader(this.prog, this.vshader);
		gl.detachShader(this.prog, this.fshader);

		this.attached = false;
	};

	this.updatevarinfo = function()
	{
		this.ainfo = getattribinfo(this.prog);
		this.uinfo = getuniforminfo(this.prog);
	};

	/* This method sets or updates the GLSL source for this program's
	 * vertex and fragment shades, then tries to (re-)compile the
	 * shaders, (re-)link the program, and retrieve information about
	 * the vertex attributes and uniform variables declared in it.
	 *
	 * Returns an array [ok, vslog, fslog, prlog] where:
	 *   ok = whether shader compilation and program linking succeeded
	 *   vslog = vertex shader info log (compiler messages)
	 *   fslog = fragment shader info log
	 *   prlog = program info log
	 */
	this.sources = function(vsrc, fsrc)
	{
		var vslog;
		var fslog;
		var prlog;

		this.detach();

		// Compile the new shaders.
		gl.shaderSource(this.vshader, vsrc);
		gl.compileShader(this.vshader);
		vslog = gl.getShaderInfoLog(this.vshader);

		gl.shaderSource(this.fshader, fsrc);
		gl.compileShader(this.fshader);
		fslog = gl.getShaderInfoLog(this.fshader);

		// If compilation failed, indicate this in the error log.
		prlog = "";
		if (!gl.getShaderParameter(this.vshader, gl.COMPILE_STATUS))
			prlog += "Vertex shader failed to compile.\n";
		if (!gl.getShaderParameter(this.fshader, gl.COMPILE_STATUS))
			prlog += "Fragment shader failed to compile.\n";
		if (prlog.length > 0)
			return [false, vslog, fslog, prlog];

		// Compilation succeeded; attach the shaders and link
		// the program.
		this.attach();
		gl.linkProgram(this.prog);
		prlog = gl.getProgramInfoLog(this.prog);

		// Retrieve shader variable information.
		// Updates this.ainfo and this.uinfo.
		this.updatevarinfo();

		return [true, vslog, fslog, prlog];
	};

	var progobj = this;
	this.uniforms = new Proxy({}, {
		set: function(target, name, value, receiver)
		{
			if (!progobj.uinfo[name])
			{
				console.warn("No such uniform: " + name);
				return true;
			}

			gl.useProgram(progobj.prog);
			var uniform = progobj.uinfo[name].loc;
			switch (progobj.uinfo[name].type)
			{
				case gl.INT:        gl.uniform1i(uniform, value); break;
				case gl.FLOAT:      gl.uniform1f(uniform, value); break;
				case gl.FLOAT:      gl.uniform1f(uniform, value); break;
				case gl.FLOAT_VEC2: gl.uniform2fv(uniform, value); break;
				case gl.FLOAT_VEC3: gl.uniform3fv(uniform, value); break;
				case gl.FLOAT_VEC4: gl.uniform4fv(uniform, value); break;
				case gl.FLOAT_MAT4:
					gl.uniformMatrix4fv(uniform, false, mat2array(value));
					break;
				default:
					console.warn("Unrecognized uniform type: " +
						progobj.uinfo[name].type);
			}
			return true;
		}
	});

	this.use = function()
	{
		if (!this.attached)
			throw new Error("Shaders must be compiled and linked before use");
		gl.useProgram(this.prog);
	}

	this.destroy = function()
	{
		gl.deleteProgram(this.prog);
		gl.deleteShader(this.fshader);
		gl.deleteShader(this.vshader);
		this.prog = this.vshader = this.fshader = null;
	};

	/* Arguments (all HTML text area elements):
	 *   vse  vertex shader source element
	 *   vle  vertex shader info log
	 *   fse  fragment shader source
	 *   fle  fragment shader info log
	 *   ple  program info log
	 */
	this.attachelements = function(vse, vle, fse, fle, ple)
	{
		// TODO: move?
	};
}

/* Creates a function that converts from a plain input array to an
 * output ArrayBuffer for WebGL. The created function takes the
 * input array, output DataView, and the current input and output
 * array/buffer positions as arguments. It returns a 2-element
 * array containing the new updated positions.
 *
 * Arguments:
 *   setter: the name of a DataView set method, such as "setFloat32"
 *   numelements: number of elements to convert (attribute vector length)
 *   elementsize: size of the scalar type (e.g. 4 for a float32 vector)
 *   alignment: adds padding bytes if necessary to maintain this alignment
 */
function mkconverter(setter, numelements, elementsize, alignment)
{
	return function(array, dataview, inpos, outpos)
	{
		for (var i=0; i<numelements; i++)
		{
			dataview[setter](outpos, array[inpos], BYTE_ORDER);
			inpos++;
			outpos += elementsize;
		}

		if (outpos % alignment != 0)
			outpos += alignment - (outpos % alignment);

		return [inpos, outpos];
	};
}

function enableattr(attr)
{
	gl.vertexAttribPointer.apply(gl, attr.vapargs);
	gl.enableVertexAttribArray(attr.vapargs[0]);
}

function disableattr(attr)
{
	gl.disableVertexAttribArray(attr.vapargs[0]);
}

function Attribute(inoffset, byteoffset, vapargs, converter)
{
	this.inoffset = inoffset;
	this.byteoffset = byteoffset;
	// Arguments for gl.vertexAttribPointer.
	this.vapargs = vapargs;
	// Function that converts from an input array to an output
	// ArrayBuffer for WebGL (see mkconverter above).
	this.converter = converter;
}

function Attributes(instride, bytestride, attrs)
{
	this.instride = instride;
	this.bytestride = bytestride;
	this.attrs = attrs;

	this.activate   = () => attrs.forEach(enableattr);
	this.deactivate = () => attrs.forEach(disableattr);
}

function mkattrinfo(progobj, posname, colorname, normalname)
{
	var attrs = [];
	var posinfo = progobj.ainfo[posname];
	var colorinfo = progobj.ainfo[colorname];
	var normalinfo = progobj.ainfo[normalname];
	var instride = 0;
	var inoffset = 0;
	var bytestride = 0;
	var byteoffset = 0;

	if (posinfo)    { instride += 3; bytestride += 12; }
	if (colorinfo)  { instride += 3; bytestride +=  4; }
	if (normalinfo) { instride += 3; bytestride +=  8; }

	if (posinfo)
	{
		attrs.push(new Attribute(
			inoffset,
			byteoffset,
			[posinfo.loc, 3, gl.FLOAT, false, bytestride, byteoffset],
			mkconverter("setFloat32", 3, 4, 4)
		));
		inoffset += 3;
		byteoffset += 12;
	}
	if (colorinfo)
	{
		attrs.push(new Attribute(
			inoffset,
			byteoffset,
			[colorinfo.loc, 3, gl.UNSIGNED_BYTE, true, bytestride, byteoffset],
			mkconverter("setUint8", 3, 1, 4)
		));
		inoffset += 3;
		byteoffset += 4; // extra byte for padding
	}
	if (normalinfo)
	{
		attrs.push(new Attribute(
			inoffset,
			byteoffset,
			[normalinfo.loc, 3, gl.SHORT, true, bytestride, byteoffset],
			mkconverter("setInt16", 3, 2, 4)
		));
	}

	return new Attributes(instride, bytestride, attrs);
}

/* Render-objects combine a shader program with vertex data.
 * A predefined set of vertex data formats are supported, which
 * consist of position and/or color and/or normal data:
 * 1. Position only: x, y, z (12 bytes total)
 * 2. Position and color: x, y, z, r, g, b, <1> (16 bytes)
 * 3. Position and normal: x, y, z, nx, ny, nz, <2> (20 bytes)
 * 3. Position, color, and normal:
 *      x, y, z, r, g, b, <1>, nx, ny, nz, <2> (24 bytes)
 *
 * x, y, and z are 32-bit floats. r, g, and b are unsigned bytes.
 * nx, ny, and nz are 16-bit signed shorts. <n> is n bytes of
 * padding required to align the position and normal data.
 *
 * The shader program object is used to determine which attributes
 * are active. The vertex data should be organized accordingly,
 * but the padding will be added automatically. To specify an
 * active attribute, declare it as a "varying" in the shader.
 */
function RenderObject(
	vdata,
	idata,
	progobj,
	transparent=false,
	posname=DEFAULT_VATTR_POS,
	colorname=DEFAULT_VATTR_COLOR,
	normalname=DEFAULT_VATTR_NORMAL)
{
	this.progobj = progobj;
	this.attrinfo = mkattrinfo(progobj, posname, colorname, normalname);

	[this.vbuf, this.vcount] = mkvbuf(vdata, this.attrinfo);

	if (idata)
	{
		[this.ibuf, this.ifmt] = mkibuf(idata);
		this.icount = idata.length;
	}

	this.transparent = !!transparent;

	// If set to an iterable object, invokes "draw" on each of
	// its elements after drawing this RenderObject.
	this.subobjs = null;

	// If set to a non-null value, overrides GL primitive type
	// being drawn (e.g. gl.POINTS).
	this.polymode = null;

	this.draw = function()
	{
		// Bind buffers.
		gl.bindBuffer(gl.ARRAY_BUFFER, this.vbuf);
		if (this.ibuf)
			gl.bindBuffer(gl.ELEMENT_ARRAY_BUFFER, this.ibuf);

		// Make the prepared shaders active.
		this.progobj.use();

		// Set up and enable shader attributes.
		this.attrinfo.activate();

		if (transparent)
		{
			gl.depthMask(false);
			gl.enable(gl.BLEND);
		}

		var curpolymode = this.polymode != null ? this.polymode : polymode;

		// Draw the object.
		if (this.ibuf)
			gl.drawElements(curpolymode, this.icount, this.ifmt, 0);
		else
			gl.drawArrays(curpolymode, 0, this.vcount);

		// Disable shader and its attributes.
		this.attrinfo.deactivate();
		gl.useProgram(null);

		if (transparent)
		{
			gl.disable(gl.BLEND);
			gl.depthMask(true);
		}

		if (this.subobjs && this.subobjs.forEach)
			this.subobjs.forEach(subobj => subobj.draw());
	};

	this.destroy = function()
	{
		if (this.ibuf)
		{
			gl.deleteBuffer(this.ibuf);
			this.ibuf = null;
		}

		gl.deleteBuffer(this.vbuf);

		// This helps a bit to guard against use after destroy,
		// and to make memory available in case this RenderObject
		// is still reachable.
		this.vbuf = null;
		this.attrinfo = null;
		this.progobj = null;
		this.draw = null;
		this.destroy = null;
	};
}

function initgl()
{
	// TODO: move element stuff to elements.js?
	gl = elements.main.getContext("webgl");

	// set options
	gl.enable(gl.DEPTH_TEST);
	gl.blendFunc(gl.SRC_ALPHA, gl.ONE_MINUS_SRC_ALPHA);
}
