/*
 * Graphic lambda calculus in WebGL. This is a clone of the "GCalc" program
 * included in the "plt-games" package from the Racket Scheme distribution.
 * The original GCalc program was written by Eli Barzilay.
 *
 * Copyright (c) 2017 Pim Goossens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

"use strict";

function Mat4()
{
	this.values = arguments.length > 0 ?
		Array.from(arguments) :
		[1, 0, 0, 0,
		 0, 1, 0, 0,
		 0, 0, 1, 0,
		 0, 0, 0, 1];
	this.mul = function(that)
	{
		var m1 = this.values;
		var m2 = that.values;
		return new Mat4(
			m1[0]*m2[0] + m1[1]*m2[4] + m1[2]*m2[8] + m1[3]*m2[12],
			m1[0]*m2[1] + m1[1]*m2[5] + m1[2]*m2[9] + m1[3]*m2[13],
			m1[0]*m2[2] + m1[1]*m2[6] + m1[2]*m2[10] + m1[3]*m2[14],
			m1[0]*m2[3] + m1[1]*m2[7] + m1[2]*m2[11] + m1[3]*m2[15],
			m1[4]*m2[0] + m1[5]*m2[4] + m1[6]*m2[8] + m1[7]*m2[12],
			m1[4]*m2[1] + m1[5]*m2[5] + m1[6]*m2[9] + m1[7]*m2[13],
			m1[4]*m2[2] + m1[5]*m2[6] + m1[6]*m2[10] + m1[7]*m2[14],
			m1[4]*m2[3] + m1[5]*m2[7] + m1[6]*m2[11] + m1[7]*m2[15],
			m1[8]*m2[0] + m1[9]*m2[4] + m1[10]*m2[8] + m1[11]*m2[12],
			m1[8]*m2[1] + m1[9]*m2[5] + m1[10]*m2[9] + m1[11]*m2[13],
			m1[8]*m2[2] + m1[9]*m2[6] + m1[10]*m2[10] + m1[11]*m2[14],
			m1[8]*m2[3] + m1[9]*m2[7] + m1[10]*m2[11] + m1[11]*m2[15],
			m1[12]*m2[0] + m1[13]*m2[4] + m1[14]*m2[8] + m1[15]*m2[12],
			m1[12]*m2[1] + m1[13]*m2[5] + m1[14]*m2[9] + m1[15]*m2[13],
			m1[12]*m2[2] + m1[13]*m2[6] + m1[14]*m2[10] + m1[15]*m2[14],
			m1[12]*m2[3] + m1[13]*m2[7] + m1[14]*m2[11] + m1[15]*m2[15]
		);
	};
	this.transpose = function()
	{
		var v = this.values;
		return new Mat4(
			v[0], v[4],  v[8], v[12],
			v[1], v[5],  v[9], v[13],
			v[2], v[6], v[10], v[14],
			v[3], v[7], v[11], v[15]
		);
	};
	this.rotate = function(angle, x, y, z)
	{
		var mag2 = x*x + y*y + z*z;
		// First normalize the axis vector if possible/necessary.
		if (mag2 != 0 && mag2-1.0+.00001 > .00001)
		{
			mag2 = Math.sqrt(mag2);
			x /= mag2;
			y /= mag2;
			z /= mag2;
		}
		var c = Math.cos(angle);
		var s = Math.sin(angle);
		return this.mul(new Mat4(
			x*x*(1-c)+c,   x*y*(1-c)-z*s, x*z*(1-c)+y*s, 0,
			x*y*(1-c)+z*s, y*y*(1-c)+c,   y*z*(1-c)-x*s, 0,
			x*z*(1-c)-y*s, y*z*(1-c)+x*s, z*z*(1-c)+c,   0,
			            0,             0,             0, 1
		));
	};
	this.scale = function(x, y, z)
	{
		return this.mul(new Mat4(
			x, 0, 0, 0,
			0, y, 0, 0,
			0, 0, z, 0,
			0, 0, 0, 1
		));
	};
	this.translate = function(x, y, z)
	{
		return this.mul(new Mat4(
			1, 0, 0, x,
			0, 1, 0, y,
			0, 0, 1, z,
			0, 0, 0, 1
		));
	};
	this.frustum = function(l, r, b, t, n, f)
	{
		return this.mul(new Mat4(
			2*n/(r-l),         0,  (r+l)/(r-l),              0,
			        0, 2*n/(t-b),  (t+b)/(t-b),              0,
			        0,         0, -(f+n)/(f-n), -(2*f*n)/(f-n),
			        0,         0,           -1,              0
		));
	};
	this.perspective = function(angle, aspect, near, far)
	{
		var halfwidth = near * Math.tan(angle/2.0);
		return this.frustum(
			-aspect * halfwidth,
			 aspect * halfwidth,
			-halfwidth,
			 halfwidth,
			near,
			far
		);
	};
};
